/* ********************************************************************************
 *  keyboard module for use with AVR series microcontrollers from Atmel           *
 ******************************************************************************** */
 
/*
    Copyright (C) 2005  Erik Hggstrm <xpress@xpress.mine.nu>

    Copyright notice:

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <avr/io.h>
//#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "keyboard.conf.h"
#include "keyboard.h"


#include "keyboardtest.h"
/* ********************************************************************************
 *  DEFINES                                                                       *
 ******************************************************************************** */

// We want to set the clock pin from knowing the SIG_KEYBOARD interrupt
#define __AVR_ATmega644p__
#if defined(__AVR_ATmega8__) || defined(__AVR_ATmega16__) || defined(__AVR_ATmega32__) 
	#define KEYBOARD_CLOCK_DDR		DDRD
	#define KEYBOARD_CLOCK_PORT		PORTD
	#define KEYBOARD_CLOCK_PIN		PIND
	#if (SIG_KEYBOARD==SIG_INTERRUPT0)	
		#define KEYBOARD_CLOCK_P			PD2
	#elif (SIG_KEYBOARD==SIG_INTERRUPT1)
		#define KEYBOARD_CLOCK_P			PD3
	#else
		#error "Unknown interrupt for this architecture."
	#endif
#elif defined(__AVR_ATmega64__) || defined(__AVR_ATmega128__) || defined(__AVR_ATmega644p__)
	#if (SIG_KEYBOARD==SIG_INTERRUPT0) || (SIG_KEYBOARD==SIG_INTERRUPT1) || (SIG_KEYBOARD==SIG_INTERRUPT2) || (SIG_KEYBOARD==SIG_INTERRUPT3)
		#define KEYBOARD_CLOCK_DDR		DDRD
		#define KEYBOARD_CLOCK_PORT		PORTD
		#define KEYBOARD_CLOCK_PIN		PIND
		#if (SIG_KEYBOARD==SIG_INTERRUPT0)	
			#define KEYBOARD_CLOCK_P			PD0
		#elif (SIG_KEYBOARD==SIG_INTERRUPT1)
			#define KEYBOARD_CLOCK_P			PD1
		#elif (SIG_KEYBOARD==SIG_INTERRUPT2)
			#define KEYBOARD_CLOCK_P			PD2
		#elif (SIG_KEYBOARD==SIG_INTERRUPT3)
			#define KEYBOARD_CLOCK_P			PD3
		#endif
	#elif (SIG_KEYBOARD==SIG_INTERRUPT4) || (SIG_KEYBOARD==SIG_INTERRUPT5) || (SIG_KEYBOARD==SIG_INTERRUPT6) || (SIG_KEYBOARD==SIG_INTERRUPT7)
		#define KEYBOARD_CLOCK_DDR		DDRE
		#define KEYBOARD_CLOCK_PORT		PORTE
		#define KEYBOARD_CLOCK_PIN		PINE
		#if (SIG_KEYBOARD==SIG_INTERRUPT4)	
			#define KEYBOARD_CLOCK_P			PE4
		#elif (SIG_KEYBOARD==SIG_INTERRUPT5)
			#define KEYBOARD_CLOCK_P			PE5
		#elif (SIG_KEYBOARD==SIG_INTERRUPT6)
			#define KEYBOARD_CLOCK_P			PE6
		#elif (SIG_KEYBOARD==SIG_INTERRUPT7)
			#define KEYBOARD_CLOCK_P			PE7
		#endif
	#else
		#error "Unknown interrupt for this architecture."
	#endif
#else
	#error "Unknown architecture, this module currently supports"
	#error "mega8,mega16,mega32,mega64,mega128,mega644p"
	#error "minor changes might be nessesary to support your device."
#endif

#define _kb_read_io() (KEYBOARD_DATA_PIN&(1<<KEYBOARD_DATA_P))

#define enterCS() {}
#define exitCS() {}

#define KB_KC_LSHIFT		0x12
#define KB_KC_RSHIFT		0x59
#define KB_KC_CAPSLOCK		0x58

#define KB_EXT_STATUS_LSHIFT						0x01
#define KB_EXT_STATUS_RSHIFT						0x02
#define KB_EXT_STATUS_CAPSLOCK						0x04
#define KB_EXT_STATUS_CAPSLOCK_RELEASE				0x08



/* ********************************************************************************
 *  EXTERNAL VARIABLES                                                            *
 ******************************************************************************** */

extern prog_uint8_t _kb_keymap_unshifted[];
extern prog_uint8_t _kb_keymap_shifted[];

/* ********************************************************************************
 *  COMMON VARIABLES                                                              *
 ******************************************************************************** */
uint8_t _kb_bitcount;
uint8_t _kb_err;

uint8_t  get_kb_bitcount() {
	return _kb_bitcount;
}
uint8_t _kb_buf_write_pos;
uint8_t _kb_buf_read_pos;
uint8_t _kb_buf[KEYBOARD_BUF_SIZE];
uint16_t _kb_ext_status; // Separates L and R shift etc.
uint16_t _kb_status; // Shift, Ctrl, Alt, Alt Gr, Num Lock

/* ********************************************************************************
 *  PROTOTYPES                                                                    *
 ******************************************************************************** */

void _kb_buf_put(uint8_t ch);
uint8_t _kb_buf_get(void);
uint8_t _kb_buf_count(void);

/* ********************************************************************************
 *  FUNCTIONS                                                                     *
 ******************************************************************************** */

void kb_init(void) {
	uint8_t i;
	
	// Set I/O pins
	KEYBOARD_CLOCK_DDR&=~(1<<KEYBOARD_CLOCK_P);	// Set as input
	KEYBOARD_CLOCK_PORT|=(1<<KEYBOARD_CLOCK_P);	// Enable pull-up
	KEYBOARD_DATA_DDR&=~(1<<KEYBOARD_DATA_P);		// Set as input
	KEYBOARD_DATA_PORT|=(1<<KEYBOARD_DATA_P);		// Enable pull-up
	
	// Init buffer
	_kb_buf_write_pos=0;
	_kb_buf_read_pos=0;
	// Fill buffer with 0's
	for(i=0;i<KEYBOARD_BUF_SIZE;i++) _kb_buf[i]=0;
	
	// Reset variables for handler
	_kb_bitcount=0;
	_kb_err=0;
	
	// Init interrupt
	// Need to chage GICR for EICR and possible check the MCUCR flags as well
	//~ #if (SIG_KEYBOARD==SIG_INTERRUPT0)
		//~ MCUCR=(MCUCR&0xFC)|(0x02<<ISC00);		// Set falling edge
		//~ //GICR=(GICR&0xBF)|(0x01<<INT0);			// Enable interrupt
	//~ #elif (SIG_KEYBOARD==SIG_INTERRUPT1)
		//~ MCUCR=(MCUCR&0xF3)|(0x02<<ISC10);		// Set falling edge
		//~ GICR=(GICR&0x7F)|(0x01<<INT1);			// Enable interrupt
	//~ #else
		//~ #error "Unknown interrupt, this module currently supports int0 and int1"
	//~ #endif
}


uint8_t buf[20];

//~ SIGNAL(SIG_KEYBOARD) {
void callInt() {
	static uint8_t _kb_tmpdata;
	
	if(_kb_bitcount==0) {
		// Start bit should always be 0
		if(_kb_read_io()) _kb_err|=0x01;
	} else if((_kb_bitcount>0)&&(_kb_bitcount<9)) {
		// Shift tmpdata
		_kb_tmpdata=_kb_tmpdata>>1;
		// Set new bit
		if(_kb_read_io()) _kb_tmpdata|=0x80;
	} else if(_kb_bitcount==9) {
		// Parity bit
		//err|=0x02
	} else if(_kb_bitcount==10) {
		// Stop bit should always be 1
		if(!_kb_read_io()) _kb_err|=0x04;
			
		// Copy byte to buffer
		//if (_kb_tmpdata != 0Xff) // added ny ND 
			_kb_buf_put(_kb_tmpdata);
				
		//~ zf_myprint(_kb_tmpdata);
		//~ zf_myprint2(_kb_buf_write_pos);
		//~ zf_myprint2(_kb_buf_read_pos);
	}
	
	//Update bit count
	_kb_bitcount++;

	if(_kb_bitcount>=11) {
		// Prepare for new frame
		_kb_bitcount=0;
		//_kb_tmpdata=0;
		_kb_err=0;
	}
}

uint16_t kb_getDecode(void) {
	//return kb_decode(kb_get(), 0);
	return kb_get();
}

uint16_t kb_get(void) {
	uint8_t tmp;
	while(1) {
		tmp=_kb_buf_get();
		//~ zf_myprint3(999);
		if(tmp==0xF0) {
			// Break code
			return(0xF000|_kb_buf_get());
		} else if(tmp==0xE0) {
			// Extended key code
			tmp=_kb_buf_get();
			if(tmp==0xF0) {
				// Extended break code
				// Skip next char
				_kb_buf_get();
			} else {
				// Extended make code
				return(0xE000|tmp);
			}
		} else if(tmp==0xE1) {
			// Special key PAUSE
			// [E1,14,77,E1,F0,14,F0,77]
			for(tmp=0;tmp<7;tmp++) {
				_kb_buf_get();
			}
			return(0xE114);
		} else {
			// Normal make code
			return(tmp);
		}
	}
}

uint8_t kb_decode(uint16_t code,uint8_t shift) {
	uint16_t tmpaddr;
	uint8_t tmpcode;
	
	// Select codepage
	if(_kb_status&KB_STATUS_SHIFT) {
		tmpaddr=(uint16_t)_kb_keymap_shifted;
	} else {
		tmpaddr=(uint16_t)_kb_keymap_unshifted;
	}
	
	// Find code in codepage
	tmpcode=pgm_read_byte(tmpaddr); 
	
	while(!((tmpcode==code)||(tmpcode==0))) {
		tmpaddr=tmpaddr+2;
		tmpcode=pgm_read_byte(tmpaddr);
	}
	
	// Get ascii
	tmpcode=pgm_read_byte(tmpaddr+1);
	
	// If alpha and capslock
	if(_kb_status&KB_STATUS_CAPSLOCK) {
		if((tmpcode>='a')&&(tmpcode<='z')) {
			tmpcode-=32;
		} else if((tmpcode>='A')&&(tmpcode<='Z')) {
			tmpcode+=32;
		}
	}
	
	// Return ascii
	if(tmpcode!=0) return(tmpcode);
	
	// Code was not ascii, handle shift and some other things
	if((code&0xFF00)==0xF000) {
		// Break code
		tmpcode=(code&0x00FF);
		if(tmpcode==KB_KC_LSHIFT) {
			_kb_ext_status&=~KB_EXT_STATUS_LSHIFT;
		} else if(tmpcode==KB_KC_RSHIFT) {
			_kb_ext_status&=~KB_EXT_STATUS_RSHIFT;
		} else if(tmpcode==KB_KC_CAPSLOCK) {
			_kb_ext_status|=KB_EXT_STATUS_CAPSLOCK_RELEASE;
		}
	} else {
		// Make code
		if(code==KB_KC_LSHIFT) {
			_kb_ext_status|=KB_EXT_STATUS_LSHIFT;
		} else if(code==KB_KC_RSHIFT) {
			_kb_ext_status|=KB_EXT_STATUS_RSHIFT;
		} else if(code==KB_KC_CAPSLOCK) {
			if(_kb_ext_status&KB_EXT_STATUS_CAPSLOCK_RELEASE) {
				_kb_ext_status&=~KB_EXT_STATUS_CAPSLOCK_RELEASE;
				_kb_ext_status^=KB_EXT_STATUS_CAPSLOCK;
			}
		}
	}
	if(_kb_ext_status&(KB_EXT_STATUS_RSHIFT|KB_EXT_STATUS_LSHIFT)) {
		_kb_status|=KB_STATUS_SHIFT;
	} else {
		_kb_status&=~KB_STATUS_SHIFT;
	}
	if(_kb_ext_status&(KB_EXT_STATUS_CAPSLOCK)) {
		_kb_status|=KB_STATUS_CAPSLOCK;
	} else {
		_kb_status&=~KB_STATUS_CAPSLOCK;
	}
	
	return(0);
}

void kb_put(uint8_t data) {
	/* Not implemented, will be used for keyboard LED's */
}

uint16_t kb_getStatus(void) {
	return(_kb_status);
}

uint8_t kb_buf_count(void) {
	return _kb_buf_count();
}
uint8_t _kb_buf_count(void) {
	
	enterCS();
	if(_kb_buf_write_pos>=_kb_buf_read_pos) {
		return(_kb_buf_write_pos-_kb_buf_read_pos);
	} else {
		return(_kb_buf_write_pos+KEYBOARD_BUF_SIZE-_kb_buf_read_pos);
	}
	
	exitCS();
}

void _kb_buf_put(uint8_t ch) {
	
	enterCS();
	// Check if bus is full
	if(_kb_buf_count()<(KEYBOARD_BUF_SIZE-1)) {
		// Add char to buf
		_kb_buf[_kb_buf_write_pos]=ch;
		//~ zf_myprint2(_kb_buf_write_pos);
		//~ zf_myprint2(_kb_buf_read_pos);
		// Increase write pos
		if(++_kb_buf_write_pos>=KEYBOARD_BUF_SIZE) _kb_buf_write_pos=0;
		//~ zf_myprint2(_kb_buf_count());
	}
	
	exitCS();
}

uint8_t _kb_buf_get(void) {
	uint8_t tmp;
	
	// If buf is empty wait for char
	//~ zf_myprint3(_kb_buf_count());
	while(_kb_buf_count()==0) {
		//~ zf_myprint3(_kb_buf_count());
		zf_myDelay();
	}
	// Read char from buf
	tmp=_kb_buf[_kb_buf_read_pos];
	// Increase read pos
	if(++_kb_buf_read_pos>=KEYBOARD_BUF_SIZE) _kb_buf_read_pos=0;
	// Return char
	return(tmp);
}
