#include "keyboardtest.h"

#define HEAP_LIMIT			(uint8_t *)0
#define HEAP_SIZE			512

PersistData sysData EEMEM = { 0, ZX_TRUE };

// default Com1 speed
#define DEF_COM1_SPEED		19200

// transmit and receive queues for Com1
uint8_t txQueue[25] NO_INIT;
uint8_t rxQueue[50] NO_INIT;

static uint8_t zv_int0TaskStack[100];
static uint8_t zv_kbReadTask[110];

ZX_INIT_FUNC(init_keyboardtest)
{
	initQueue((Queue *)txQueue, sizeof(txQueue));
	initQueue((Queue *)rxQueue, sizeof(rxQueue));
	zxInit(BAUD_FACTOR_2X(DEF_COM1_SPEED));
#if defined(ENABLE_SW_UART)
	enableSerialSW();
#endif

}

int
main(void)
{
#if defined(ZX_INIT)
	ZX_INIT;
#endif
	taskStartMain(zf_Main, HEAP_LIMIT, HEAP_SIZE);
	for(;;);
}

void
zf_Main(void)
{
	initHeap();

	kb_init();
	taskStart(zf_int0Task, zv_int0TaskStack, 100, 0);
	taskStart(zf_kbRead, zv_kbReadTask, 110, 0);
	while (1)
	{
		zf_flashLedPattern();
	}
}

void
zf_kbRead(void)
{
	uint16_t zv_keyVal;

	{
		uint8_t zv_bitCount;

		while (1)
		{
			zv_keyVal = kb_get();
			zv_bitCount = get_kb_bitcount();
			outStr(STRTBL(0));
			outStr(cstrHex(zv_keyVal, 2));
			outStr(STRTBL(5));
			outStr(byteToString(zv_bitCount));
			sendCRLF();
			taskDelay(0.1);
		}
	}
}

void
zf_int0Task(void)
{
	while (1)
	{
		waitForInterrupt(24, 0);
		callInt();
	}
}

void
zf_myDelay(void)
{
	taskDelay(0.1);
}

void
zf_flashLedPattern(void)
{
	putPin(26, 0);
	putPin(25, 1);
	taskDelay(0.5);
	putPin(26, 1);
	putPin(25, 0);
	taskDelay(0.5);
}

// string table
char stringTbl[] PROGMEM =
{
	// STRTBL(0) - "key="
	0x04, 0x6b, 0x65, 0x79, 0x3d, 

	// STRTBL(5) - "bitcount="
	0x09, 0x62, 0x69, 0x74, 0x63, 0x6f, 0x75, 0x6e, 0x74, 0x3d, 
};
